#include "fill.h"

/* local score of rotated piece p at position x,y */
int local_score(int grid[WIDTH][HEIGHT], int p, int x, int y)
{
	int s = 0;
	s += (x == 0          ? rotations[p][left]  == 0 : rotations[p][left]  == rotations[grid[x-1][y]][right]);
	s += (x == WIDTH - 1  ? rotations[p][right] == 0 : rotations[p][right] == rotations[grid[x+1][y]][left] );
	s += (y == 0          ? rotations[p][up]    == 0 : rotations[p][up]    == rotations[grid[x][y-1]][down] );
	s += (y == HEIGHT - 1 ? rotations[p][down]  == 0 : rotations[p][down]  == rotations[grid[x][y+1]][up]   );
	return s;
}

char swapdone[16][16][16][16];

/* return the number of points gained by one of the best swappings */
int swap(int grid[WIDTH][HEIGHT])
{
	int i1, j1, s1, i2, j2, s2, s, p1, p2, r1, r2, n1, n2;
	int found = 0, best = -99, bi1 = 0, bj1 = 0, bp1 = 0, br1 = 0, bi2 = 0, bj2 = 0, bp2 = 0, br2 = 0;
	
	/* for all pieces who are not hints */
	for (i1 = 0; i1 < WIDTH; i1++) for (j1 = 0; j1 < HEIGHT; j1++) if (!ishint[i1][j1])
	{
		/* compute the score generated by the first piece */
		s1 = local_score(grid, grid[i1][j1], i1, j1);
		/* if it is not 4 we may optimize it */
		//if (s1 < 4)
		{
			/* try all swaps */
			for (i2 = 0; i2 < WIDTH; i2++) for (j2 = j1; j2 < HEIGHT; j2++) if (!ishint[i2][j2])
			/* don't swap the same or adjacent */
			if (abs(i1-i2) + abs(j1-j2) > 1 && !swapdone[i1][j1][i2][j2])
			{
				/* compute the score generated by the second piece */
				s2 = local_score(grid, grid[i2][j2], i2, j2);
				/* if it is not 4 we may optimize it */
				//if (s2 < 4)
				if (s1 + s2 < 8)
				{
					/* compute the total score */
					s = s1 + s2;
					p1 = grid[i1][j1] / 4;
					p2 = grid[i2][j2] / 4;
					/* try all rotations */
					for (r1 = 0; r1 <= 3; r1++) for (r2 = 0; r2 <= 3; r2++)
					{
						n1 = local_score(grid, p1 * 4 + r1, i2, j2);
						n2 = local_score(grid, p2 * 4 + r2, i1, j1);
						if (n1 + n2 - s > best)
						{
							best = n1 + n2 - s;
							found = 1;
						}
						if (n1 + n2 - s >= best && rand()%found<1)
						{
							bi1 = i1;
							bj1 = j1;
							bp1 = p1;
							br1 = r1;
							bi2 = i2;
							bj2 = j2;
							bp2 = p2;
							br2 = r2;
							found++;
						}
					}
				}
			}
		}
	}
	/*
	 * even if we loose some points, ie best < 0
	 * we try, we might gain more later
	 */
	swapdone[bi1][bj1][bi2][bj2] = 1;
	swapdone[bi2][bj2][bi1][bj1] = 1;
	grid[bi1][bj1] = bp2 * 4 + br2;
	grid[bi2][bj2] = bp1 * 4 + br1;
	pps+=2;
	pps_swap+=2;
	pps_total+=2;
	pps_restart+=2;
	return best;
}

/* try to optimize a grid by doing some swaps */
void optimize(int score)
{
	int test[WIDTH][HEIGHT], optimized[WIDTH][HEIGHT];
	int delta, bonus = 0, i;
	
	/* try several times */
	for (i = 0; i < coeff4; i++)
	{
		bzero(swapdone, sizeof(swapdone));
		delta = 0;
		memcpy(test, curgrid, sizeof(test));
		/* swap until blocked */
		for (;;)
		{
			delta += swap(test);
			/* we lost too much points, useless to continue */
			if (delta < -coeff5) break;
			if (delta > bonus)
			{
				bonus = delta;
				memcpy(optimized, test, sizeof(optimized));
			}
		}
	}
	#if DEBUG > 1
	LOG(stdout, "Optimize: %d+%d\n", score, bonus);
	#endif
	
	/* if the result of optimize if good, we save the new best grid */
	if (bonus > 0 && score + bonus >= best) save(optimized, score + bonus);
}

int is_ok(piece p, piece c)
{
	/* if there are other constraints and not matching */
	if (c[right] != ANY && p[right] != c[right]) return 0;
	if (c[down] != ANY && p[down] != c[down]) return 0;
	
	/* if there are no constraints but border piece */
	if (c[right] == ANY && p[right] == 0) return 0;
	if (c[down] == ANY && p[down] == 0) return 0;
	
	/* this piece is ok */
	return 1;
}

void compute_constraints(piece c, int x, int y)
{
	c[left] = x > 0 ? rotations[curgrid[x-1][y]][right] : 0;
	c[up] = y > 0 ? rotations[curgrid[x][y-1]][down] : 0;
	c[right] = x < WIDTH - 1 ? rotations[curgrid[x+1][y]][left] : 0;
	c[down] = y < HEIGHT - 1 ? rotations[curgrid[x][y+1]][up] : 0;
}

void compute_position(int step, int *rx, int *ry)
{
	/* simple one */
	/*
	*rx = step % WIDTH;
	*ry = step / WIDTH;
	*/
	
	/* nicer one */
	*rx = x[step];
	*ry = y[step];
}

void fill(int step, int score, int skipped)
{
	int x, y, i, p;
	int ok = 0;
	piece c;
	
	/* should we restart? */
	if (coeff6 && pps_restart > coeff6)
		return;
	
	/* is it better? */
	if (score >= best) save(curgrid, score);
	
	/* did we fill the grid? */
	if (step == WIDTH * HEIGHT)
	{
		#ifndef _PSP_FW_VERSION
		if (!benchmark && score > best - coeff7) optimize(score);
		#endif
		return;
	}
	
	/* should we draw the gui? */
	#ifdef _PSP_FW_VERSION
	if (pps % 10000 == 0) drawnow = 1;
	if (pps >= 1000000 && benchmark) benchmark_timer();
	#endif
	if (gui && drawnow) draw();
	if (minigui && drawnow_mini) draw_mini();
	
	/* compute the positon */
	compute_position(step, &x, &y);
	
	/* is it a hint? */
	if (!benchmark && ishint[x][y])
	{
		compute_constraints(c, x, y);
		i = (c[up]==rotations[curgrid[x][y]][up]) + (c[left]==rotations[curgrid[x][y]][left]);
		fill(step+1, score+i, skipped+2-i);
		return;
	}
	
	/* increment stat counter for this step */
	pps++;
	if (skipped) pps_optim++;
	pps_total++;
	pps_restart++;
	
	/* compute constraints */
	compute_constraints(c, x, y);
	
	/* try all the pieces */
	for (i = 0; (p = havecolors[(int)c[up]][(int)c[left]][i]); i++)
		
		/* if this specific piece is eligible for this place */
		if (!onboard[p/4])
		
		/* test if the piece is ok at this place */
		if (is_ok(rotations[p], c))
		{
			/* ok we place it */
			curgrid[x][y] = p;
			onboard[p/4] = 1;
			/* and we continue */
			fill(step+1, score+2, skipped);
			/* we came back, we can remove it */
			curgrid[x][y] = 0;
			onboard[p/4] = 0;
			ok = 1;
		}
	
	/* try to skip but not too much... */
	if (!ok)
	if (!benchmark)
	if (skipped <= maxskip[step])
	if (skipped <= 480 - best)
	if (skipped <= coeff8)
	{
		for (i = 0; (p = havecolors[ANY][(int)c[left]][i]); i++)
		if (!onboard[p/4])
		if ((x == WIDTH - 1) ^ (rotations[p][right] != 0))
		if ((y == HEIGHT - 1) ^ (rotations[p][down] != 0))
		{
			curgrid[x][y] = p;
			onboard[p/4] = 1;
			fill(step+1, score+1, skipped+1);
			curgrid[x][y] = 0;
			onboard[p/4] = 0;
			ok = 1;
		}
		for (i = 0; (p = havecolors[(int)c[up]][ANY][i]); i++)
		if (!onboard[p/4])
		if ((x == WIDTH - 1) ^ (rotations[p][right] != 0))
		if ((y == HEIGHT - 1) ^ (rotations[p][down] != 0))
		{
			curgrid[x][y] = p;
			onboard[p/4] = 1;
			fill(step+1, score+1, skipped+1);
			curgrid[x][y] = 0;
			onboard[p/4] = 0;
			ok = 1;
		}
	}
}
